@@ -6,16 +6,19 @@
 # redistribute it and/or modify it under the same terms as Perl
 # itself.
 #
-# $Id: Emaul.pm,v 1.2 1996/07/16 04:47:18 kjj Exp $
+# $Id: Emaul.pm,v 1.3 1996/08/03 17:32:21 kjj Exp $
 
 package Mail::Folder::Emaul;
 @ISA = qw(Mail::Folder);
+use Mail::Folder;
+
+Mail::Folder::register_folder_type(Mail::Folder::Emaul, 'emaul');
 
 =head1 NAME
 
 Mail::Folder::Emaul - An Emaul folder interface for Mail::Folder.
 
-I<B<WARNING: This code is in alpha release.> Expect the interface to
+B<WARNING: This code is in alpha release. Expect the interface to
 change.>
 
 =head1 SYNOPSYS
@@ -33,9 +36,6 @@ directories for folders and numerically-named files for the individual
 mail messages.  The current message for a particular folder is stored
 in a file C<.current_msg> in the folder directory.
 
-Many of B<mh>'s more useful features like sequences aren't
-implemented, but what the hey...
-
 =cut
 
 use Mail::Folder;
@@ -43,12 +43,17 @@ use Mail::Internet;
 use Carp;
 use vars qw($VERSION);
 
-$VERSION = "0.02";
+$VERSION = "0.03";
 
 =head1 METHODS
 
 =head2 open($folder_name)
 
+Populates the C<Mail::Folder> object with information about the
+folder.
+
+=over 2
+
 =item * Call the superclass C<open> method
 
 =item * For every message file in the C<$folder_name> directory, add the
@@ -57,44 +62,43 @@ message_number to the object's list of messages.
 =item * Load the contents of C<$folder_dir/.current_msg> into
 C<$self-E<gt>{Current}>.
 
+=back
+
 =cut
 
 sub open {
   my $self = shift;
-  my $file = shift;
-  
+  my $foldername = shift;
+
   my($current_message, $message);
   local(*FILE);
   
-  return(0) unless $self->SUPER::open($file);
+  return(0) unless $self->SUPER::open($foldername);
+
+  $self->set_readonly if (! -w $foldername);
+
+  map($self->remember_message($_), get_folder_msgs($foldername));
+  $self->sort_message_list;
+
+  $self->current_message(load_current_msg($foldername));
+  $self->load_message_labels;
 
-  map($self->remember_message($_), get_folder_msgs($self->foldername()));
-  $self->sort_message_list();
-  
-  if (open(FILE, $self->foldername() . "/.current_msg")) {
-    $current_message = <FILE>;
-    close(FILE);
-    chomp($current_message);
-    croak("non-numeric content in " . $self->foldername() . "/.current_msg")
-      if ($current_message !~ /^\d+$/);
-    $self->current_message($current_message);
-  }
   return(1);
 }
 
-=head2 close()
+=head2 close
 
 Does nothing except call the C<close> method of the superclass.
 
 =cut
 
-sub close {
-  my $self = shift;
-  
-  return($self->SUPER::close());
-}
+sub close { return($_[0]->SUPER::close); }
 
-=head2 sync()
+=head2 sync
+
+Flushes any pending changes out to the original folder.
+
+=over 2
 
 =item * Call the superclass C<sync> method
 
@@ -113,41 +117,44 @@ around in the Mail::Folder object.
 
 =item * Return the number of new messages found.
 
+=back
+
 =cut
 
 sub sync {
   my $self = shift;
   
-  my $current_message = $self->current_message();
+  my $current_message = $self->current_message;
   my $qty_new_messages = 0;
-  my $msg;
-  my @msgs;
-  local(*FILE);
+  my @deletes = $self->select_label('delete');
+  my $foldername = $self->foldername;
   
-  return(-1) if ($self->SUPER::sync() == -1);
+  return(-1) if ($self->SUPER::sync == -1);
 
-  foreach $msg (get_folder_msgs($self->foldername())) {
-    if (!defined($self->{Messages}{$msg})) {
-      $self->remember_message($msg);
+  for (get_folder_msgs($foldername)) {
+    if (!defined($self->{Messages}{$_})) {
+      $self->remember_message($_);
       $qty_new_messages++;
     }
   }
-  
-  @msgs = map {$self->foldername() . "/$_"} $self->list_deletes();
-  unlink(@msgs);
-  map {$self->forget_message($_)} $self->list_deletes();
-  $self->clear_deletes();
-  $self->sort_message_list();
-  
-  open(FILE, ">" . $self->foldername() . "/.current_msg") ||
-    croak("can't write " . $self->foldername() . "/.current_msg: $!");
-  print(FILE "$current_message\n");
-  close(FILE);
+
+  if (!$self->is_readonly && @deletes) {
+    unlink(map { "$foldername/$_" } @deletes);
+    map {$self->forget_message($_)} @deletes;
+    $self->clear_label('delete');
+  }
+
+  $self->sort_message_list;
+
+  if (!$self->is_readonly) {
+    store_current_msg($foldername, $current_message);
+    $self->store_message_labels($foldername);
+  }
   
   return($qty_new_messages);
 }
 
-=head2 pack()
+=head2 pack
 
 Calls the superclass C<pack> method.
 
@@ -170,26 +177,24 @@ sub pack {
   my $self = shift;
   
   my($newmsg) = 0;
-  my($folder) = $self->foldername();
-  my($current_message) = $self->current_message();
-  my $msg;
+  my($folder) = $self->foldername;
+  my($current_message) = $self->current_message;
   
-  return(0) unless $self->SUPER::pack();
-  
-  foreach $msg ($self->message_list()) {
+  return(0) if (!$self->SUPER::pack || $self->is_readonly);
+
+  for ($self->message_list) {
     $newmsg++;
-    if ($msg > $newmsg) {
-      return(0) if (!rename("$folder/$msg", "$folder/$newmsg"));
-      if (-e "$folder/,$msg") {
-	return(0) if (!rename("$folder/,$msg", "$folder/,$newmsg"));
-      }
-      $self->current_message($newmsg) if ($msg == $current_message);
+    if ($_ > $newmsg) {
+      return(0) if (!rename("$folder/$_", "$folder/$newmsg") ||
+		    (-e "$folder/,$_" &&
+		     !rename("$folder/,$_", "$folder/,$newmsg")));
+      $self->current_message($newmsg) if ($_ == $current_message);
       $self->remember_message($newmsg);
-      $self->cache_header($newmsg, $self->{Headers}{$msg});
-      $self->forget_message($msg);
+      $self->cache_header($newmsg, $self->{Messages}{$_}{Header});
+      $self->forget_message($_);
     }
   }
-  $self->sort_message_list();
+  $self->sort_message_list;
   return(1);
 }
 
@@ -212,9 +217,9 @@ sub get_message {
   return(undef) unless $self->SUPER::get_message($key);
 
   if (defined($self->{Messages}{$key})) {
-    open(FILE, $self->foldername() . "/$key") ||
-      croak("can't open " . $self->foldername() . "/$key: $!");
-    $message = Mail::Internet->new(<FILE>);
+    open(FILE, $self->foldername . "/$key") ||
+      croak("can't open " . $self->foldername . "/$key: $!");
+    $message = new Mail::Internet(<FILE>);
     close(FILE);
     return($message);
   }
@@ -223,15 +228,9 @@ sub get_message {
 
 =head2 get_header($msg_number)
 
-The C<$self-E<gt>{Headers}> associative array in a B<Mail::Folder>
-object is used to hold B<Mail::Internet> object references containing
-the headers of the mail messages in the folder.  For performance
-reasons, the header references aren't retrieved immediately by the
-C<open> method, they are retrieved by C<get_header> as needed.
-
 If the particular header has never been retrieved then C<get_header>
-loads the header of the given mail message into
-C<$self-E<gt>{Headers}{$msg_number}> and returns the object reference
+loads the header of the given mail message into a member of
+C<$self-E<gt>{Messages}{$msg_number}> and returns the object reference
 
 If the header for the given mail message has already been retrieved in
 a prior call to C<get_header>, then the cached entry is returned.
@@ -247,12 +246,12 @@ sub get_header {
   my $header;
   local(*FILE);
 
-  return($self->{Headers}{$key}) if ($self->{Headers}{$key});
+  return($self->{Messages}{$key}{Header}) if ($self->{Messages}{$key}{Header});
   
   return(undef) unless $self->SUPER::get_header($key);
 
-  if (open(FILE, $self->foldername() . "/$key")) {
-    $header = Mail::Internet->new();
+  if (open(FILE, $self->foldername . "/$key")) {
+    $header = new Mail::Internet;
     $header->read_header(\*FILE);
     close(FILE);
     $self->cache_header($key, $header);
@@ -263,6 +262,11 @@ sub get_header {
 
 =head2 append_message($message_ref)
 
+Appends the contents of the mail message contained C<$message_ref> to
+the the folder.
+
+=over 2
+
 =item * Call the superclass C<append_message> method.
 
 =item * Retrieve the highest message number in the folder
@@ -272,10 +276,12 @@ sub get_header {
 =item * Create a new mail message file in the folder with the
 contents of C<$message_ref>.
 
-Please note that, contrary to the documentation for B<Mail::Folder>,
-actually updates the real folder, rather than queueing it up for a
-subsequent sync.  The C<dup> and C<refile> methods are also
-affected. This will be fixed soon.
+=back
+
+Please note that, contrary to other documentation for B<Mail::Folder>,
+the Emaul C<append_message> method actually updates the real folder,
+rather than queueing it up for a subsequent sync.  The C<dup> and
+C<refile> methods are also affected. This will be fixed soon.
 
 =cut
 
@@ -283,29 +289,37 @@ sub append_message {
   my $self = shift;
   my $message_ref = shift;
   
-  my($message_num) = $self->last_message();
+  my($message_num) = $self->last_message;
   local(*FILE);
   
   return(0) unless $self->SUPER::append_message($message_ref);
 
   $message_num++;
-  write_message($self->foldername(), $message_num, $message_ref);
+  write_message($self->foldername, $message_num, $message_ref);
   $self->remember_message($message_num);
-  $self->cache_header($message_num, $message_ref->header());
-  $self->sort_message_list();
+  $self->cache_header($message_num, $message_ref->header);
+  $self->sort_message_list;
   return(1);
 }
 
 =head2 update_message($msg_number, $message_ref)
 
+Replaces the message pointed to by C<$msg_number> with the contents of
+the C<Mail::Internet> object reference C<$message_ref>.
+
+=over 2
+
 =item * Call the superclass C<update_message> method.
 
 =item * Replaces the contents of the given mail file with the contents of
 C<$message_ref>.
 
-Please note that, contrary to the documentation for B<Mail::Folder>,
-actually updates the real folder, rather than queueing it up for a
-subsequent sync.  This will be fixed soon.
+=back
+
+Please note that, contrary to other documentation for B<Mail::Folder>,
+the Emaul C<update_message> method actually updates the real folder,
+rather than queueing it up for a subsequent sync.  This will be fixed
+soon.
 
 =cut
 
@@ -318,10 +332,28 @@ sub update_message {
   
   return(0) unless $self->SUPER::update_message($key, $message_ref);
 
-  write_message($self->foldername(), $key, $message_ref);
+  write_message($self->foldername, $key, $message_ref);
   
   return(1);
 }
+
+=head2 create($foldername)
+
+Creates a new folder named C<$foldername>.  Returns C<0> if the folder
+already exists, otherwise returns of the folder creation was
+successful.
+
+=cut
+
+sub create {
+  my $self = shift;
+  my $foldername = shift;
+
+  return(0) if (-e $foldername);
+
+  mkdir($foldername, 0700) || croak("can't create $foldername: $!\n");
+  return(1);
+}
 ###############################################################################
 sub get_folder_msgs {
   my $folder_dir = shift;
@@ -332,8 +364,8 @@ sub get_folder_msgs {
   opendir(DIR, $folder_dir) || croak("can't open $folder_dir: $!");
   @files = grep(/^\d+$/, readdir(DIR));
   closedir(DIR);
-  
-  return(sort {$a <=> $b} @files);
+
+  return(@files);
 }
 
 sub write_message {
@@ -348,14 +380,107 @@ sub write_message {
   
   open(FILE, ">$folder_dir/$key") ||
     croak("can't write $folder_dir/$key: $!");
-  chmod(0600, "$folder_dir/.current_msg") ||
-    croak("can't chmod $folder_dir/.current_msg: $!");
+  chmod(0600, "$folder_dir/$key") ||
+    croak("can't chmod $folder_dir/$key: $!");
   $message_ref->print(\*FILE);
   close(FILE);
   
   return(1);
 }
 
+sub load_current_msg {
+  my $foldername = shift;
+  my $current_msg = 0;
+  local(*FILE);
+  
+  if (open(FILE, "$foldername/.current_msg")) {
+    $current_msg = <FILE>;
+    close(FILE);
+    chomp($current_msg);
+    croak("non-numeric content in $foldername/.current_msg")
+      if ($current_msg !~ /^\d+$/);
+  }
+  return($current_msg);
+}
+
+sub store_current_msg {
+  my $foldername = shift;
+  my $current_msg = shift;
+  
+  local(*FILE);
+
+  open(FILE, ">$foldername/.current_msg") ||
+    croak("can't write $foldername/.current_msg: $!");
+  print(FILE "$current_msg\n");
+  close(FILE);
+}
+
+sub store_message_labels {
+  my $self = shift;
+  my @alllabels = $self->list_all_labels;
+  my @labels;
+  local(*FILE);
+
+  if (@alllabels) {
+    open(FILE, ">" . $self->foldername . "/.msg_labels") ||
+      croak("can't create " . $self->foldername . "/.msg_labels: $!\n");
+    for (@alllabels) {
+      @labels = $self->select_label($_);
+      print(FILE "$_: ", collapse_select_list(@labels), "\n");
+    }
+    close(FILE);
+  }
+}
+
+sub collapse_select_list {
+  my @list = @_;
+  my @commalist;
+  my $low = $list[0];
+  my $high = $low;
+
+  for (@list) {
+    if ($_ > ($low + 1)) {
+      push(@commalist, ($low != $high) ? "$low-$high" : $low);
+      $low = $_;
+    }
+    $high = $_;
+  }
+  push(@commalist, ($low != $high) ? "$low-$high" : $low);
+  return(join(',', @commalist));
+}
+
+sub load_message_labels {
+  my $self = shift;
+
+  my %labels;
+  my($label, $value);
+  my($commachunk, $low, $high);
+  local(*FILE);
+
+  if (open(FILE, $self->foldername . "/.msg_labels")) {
+    while (<FILE>) {
+      chomp;
+      next if (/^\s*$/);
+      next if (/^\s*\#/);
+      ($label, $value) = split(/\s*:\s*/, $_, 2);
+      $labels{$label} = $value;
+      foreach $commachunk (split(',', $value)) {
+	if ($commachunk =~ /-/) {
+	  ($low, $high) = split(/-/, $commachunk, 2);
+	} else { $low = $high = $commachunk; }
+	($low <= $high) || croak("bad message spec: $low > $high: $value\n");
+	(($low =~ /^\d+$/) && ($high =~ /^\d+$/)) ||
+	  croak("bad message spec: $value\n");
+	for (; $low <= $high; $low++) {
+	  ($self->add_label($low, $label))
+	    if (defined($self->{Messages}{$low}));
+	}
+      }
+    }
+    close(FILE);
+  }
+}
+
 =head1 AUTHOR
 
 Kevin Johnson E<lt>F<kjj@primenet.com>E<gt>
